<?php
/*

  Copyright (c) 2012 Marcin Konicki <ahwayakchih@neoni.net>
  Script created for drupion.com.
 
  # About
  This file will modify `.htaccess` file to be able to check session ID of
  Ubersmith's user.

  It sets a cookie with `sha()` (or equivalent - depends on database)
  of session ID available to all subdomains in configured domain. That allows
  other sites in the same domain (they may be in different subdomains) to check
  if user is currently logged into Ubersmith.

  It also allows to call `logout.php` with `?redirect=[URL]`, to logout user
  and redirect her/him to the specified URL.

  # Installation
  1. Configure variables below and upload this file to the main directory of
     Ubersmith's installation.
  2. Call this script with `?install` option, e.g.,
     http://example.com/login_trigger.php?install from a whitelisted host.
     If install variable has value, script will answer with the same value when
     installation is completed.

  # Uninstallation
  1. Call this script with ?uninstall option, e.g.,
     http://example.com/login_trigger.php?uninstall from a whitelisted host.
     If uninstall variable has value, script will answer with the same value
     when deinstallation is completed.
  2. Delete this script from Ubersmith's directory.

  # TODO:
  1. Consider implementing "api-keys" to be used instead of or along with
     $whitelist of IP numbers.

*/

// Uncomment next two lines for debugging.
//ini_set("display_errors","2");
//ERROR_REPORTING(E_ALL);

/*
 * Change settings below to match your installation.
 */

// Required: Our cookie name where we store last valid token. It will be accessible for all subdomains.
$cname = '/* COOKIE NAME */'; // Example: "DRUPION"
// Required: Domain that keeps our subdomains. Prefixed with a dot.
//           The same domain must be used for Ubersmith and any site that wants to use automatic login.
$domain = '/* DOMAIN NAME */'; // Example: ".drupion.com"

// Optional (not needed when using Ubersmith's config.ini.php):
// Login information required to access Ubersmith's database.
$ubersmith = array(
  'driver' => '/* DATABASE TYPE */', // Example: "mysql" or "sqlite"
  'server' => '/* SERVER ADDRESS */',
  'port' => '/* SERVER PORT */',
  'database' => '/* DATABASE NAME */',
  'user' => '/* USERNAME */',
  'pass' => '/* PASSWORD */',
  // Secret key that will be passed to sha() function 
  // along with Ubersmith's session ID to generate access token.
  // Example: "do not use this example as secret key keep it unique :)"
  'secret' => '/* SECRET KEY */',
);
// Required: Read it from Ubersmith's config file. You can remove this part,
//           if database access is configured above.
$temp = parse_ini_file('../config.ini.php', true);
if (!empty($temp) && isset($temp['database']['dsn'])) {
  $temp = parse_url($temp['database']['dsn']);
  $ubersmith['driver'] = $temp['scheme'];
  $ubersmith['server'] = $temp['host'];
  $ubersmith['port'] = $temp['port'];
  $ubersmith['database'] = basename($temp['path']);
  $ubersmith['user'] = $temp['user'];
  $ubersmith['pass'] = $temp['pass'];
}

// Required: List of IP numbers that can access install, uninstall and validate functions.
$whitelist = array(
  $_SERVER['SERVER_ADDR'],
  /* WHITELISTED IPs */
);

/*
 * Change following settings only if you know what you're doing :).
 */

// Session name (name of a session cookie) used by Ubersmith.
$sname = 'UBERSID';
// File that should be overriden by our voodoo. This should be a file that is called
// at least every time when Ubersmith user is logging in and out, so our script can set
// cookies properly.
// `js/common.js.php` is called from every page, so it's a good choice.
$hookto = 'js/common.js.php';
// Mark if overriden file is a script that should be executed, e.g., when it's a PHP script.
$hookrun = true;
// Just a unique string that will mark our magic in `.htaccess` file for easy recognition.
$HTAID = 'multi-domain logins';

/*
 * You should not change anything below this line.
 */

// Prepare our magic that will be glued to `.htaccess` file.
$magic = "### START: {$HTAID} ###
RewriteEngine On
RewriteCond %{DOCUMENT_ROOT}/".basename(__FILE__)." -f
RewriteCond %{REQUEST_URI} ^/{$hookto}
RewriteRule ^(.*)$ ".basename(__FILE__)."?q=$1 [NC,QSA,L]

RewriteCond %{DOCUMENT_ROOT}/".basename(__FILE__)." -f
RewriteCond %{REQUEST_URI} ^/(logout.php|index.php|)$ [NC]
RewriteCond %{QUERY_STRING} (^|&)redirect= [NC]
RewriteRule ^(.*)$ ".basename(__FILE__)."?q=\$1 [NC,QSA,L]

RewriteCond %{DOCUMENT_ROOT}/".basename(__FILE__)." -f
RewriteCond %{HTTP_COOKIE} redirect=([^;]+) [NC]
RewriteRule ^(.*)$ ".basename(__FILE__)."?redirect_now=1&q=\$1 [NC,QSA,L]
### END: {$HTAID} ###";

function userip() {
  static $ip;
  if ($ip != NULL) return $ip;

  $proxy = NULL;

  if (isset($_SERVER["HTTP_X_FORWARDED_FOR"])) {
    if (isset($_SERVER["HTTP_CLIENT_IP"])) {
      $proxy = $_SERVER["HTTP_CLIENT_IP"];
    } else {
      $proxy = $_SERVER["REMOTE_ADDR"];
    }
    $ip = $_SERVER["HTTP_X_FORWARDED_FOR"];
  } else {
    if (isset($_SERVER["HTTP_CLIENT_IP"])) {
      $ip = $_SERVER["HTTP_CLIENT_IP"];
    } else {
      $ip = $_SERVER["REMOTE_ADDR"];
    }
  }

  return $ip;
}

// Validate if given token is currently logged in.
// If `$ip` is passed, it will be compared with IP stored with session for a match.
function validate($sid = NULL, $ip = NULL) {
  if (empty($sid)) return false;

  global $ubersmith;
  $temp = NULL;

  // Try using PDO first.
  if (class_exists('PDO')) {
    try {
      $dsn = '';
      if ($ubersmith['driver'] == 'sqlite' || $ubersmith['driver'] == 'sqlite2') {
        $dsn = $ubersmith['driver'].':'.$ubersmith['database'];
      }
      else {
        $dsn = $ubersmith['driver'].':host='.$ubersmith['server'].';dbname='.$ubersmith['database'].';'.(!empty($ubersmith['port']) ? 'port='.$ubersmith['user'] : '');
      }

      $db = new PDO($dsn, $ubersmith['user'], $ubersmith['pass']);
      if (!$db) return false;

      $safesid = $db->quote($sid, PDO::PARAM_STR);

      $tokensql = "CONCAT(s.`id`, s.`username`, ".$db->quote($ubersmith['secret'], PDO::PARAM_STR).")";
      if ($ubersmith['driver'] == 'pgsql') {
        $tokensql = "digest({$tokensql}, 'sha256')";
      }
      elseif ($ubersmith['driver'] == 'mysql' || $ubersmith['driver'] == 'mysqli') {
        $tokensql = "sha({$tokensql})";
      }
      elseif ($ubersmith['driver'] == 'sqlite') {
        $db->sqliteCreateFunction('loginSHA', 'hash', 2);
        $tokensql = "loginSHA('sha256', {$tokensql})";
      }

      $r = $db->query("SELECT s.*, p.`username` AS admin_login, p.`email` AS admin_email, p.`name` AS admin_realname, cl.`login` AS client_login, cl.`email` AS client_email, CONCAT(cl.`first`, ' ', cl.`last`) AS client_realname, co.`login` AS contact_login, CONCAT(co.`email_name`, '@', co.`email_domain`) AS contact_email, co.`real_name` AS contact_realname, {$tokensql} AS token FROM `session` s LEFT JOIN `people` p ON p.`id` = s.`admin_id` LEFT JOIN `CLIENT` cl ON cl.`clientid` = s.`client_id` LEFT JOIN `contacts` co ON co.`contact_id` = s.`contact_id` WHERE s.`id` = {$safesid} OR {$tokensql} = {$safesid} LIMIT 1");
      if (!$r) return false;

      $temp = $r->fetch(PDO::FETCH_ASSOC);
      $db = NULL;
    }
    catch (PDOException $e) {
      $temp = NULL;
    }
  }

  // Fallback to mysql, when it makes sense.
  if ($temp == NULL && function_exists('mysql_connect') && $ubersmith['driver'] == 'mysql') {
    $db = mysql_connect($ubersmith['server'], $ubersmith['user'], $ubersmith['pass']);
    if (!$db) return false;
    mysql_select_db($ubersmith['database']);

    $safesid = mysql_real_escape_string($sid);
    $tokensql = "sha(CONCAT(s.`id`, s.`username`, '".mysql_real_escape_string($ubersmith['secret'])."'))";
    $r = mysql_query("SELECT s.*, p.`username` AS admin_login, p.`email` AS admin_email, p.`name` AS admin_realname, cl.`login` AS client_login, cl.`email` AS client_email, CONCAT(cl.`first`, ' ', cl.`last`) AS client_realname, co.`login` AS contact_login, CONCAT(co.`email_name`, '@', co.`email_domain`) AS contact_email, co.`real_name` AS contact_realname, {$tokensql} AS token FROM `session` s LEFT JOIN `people` p ON p.`id` = s.`admin_id` LEFT JOIN `CLIENT` cl ON cl.`clientid` = s.`client_id` LEFT JOIN `contacts` co ON co.`contact_id` = s.`contact_id` WHERE s.`id` = '{$safesid}' OR {$tokensql} = '{$safesid}' LIMIT 1");
    if (!$r) return false;

    $temp = mysql_fetch_array($r);
    mysql_close($db);
  }

  if (!empty($temp) && !empty($temp['username'])) {
    if (!empty($ip) && $ip != $temp['ip_address']) return false;
    return $temp;
  }

  return false;
}

// Check if session is valid and user is logged in.
// Notify others if just logged in.
// Return valid token or empty string if not logged in.
function login($sid) {
  global $cname;

  $valid = validate($sid);
  if (empty($valid) || empty($valid['username'])) return '';

  if (!isset($_COOKIE[$cname]) || $_COOKIE[$cname] != $sid) {
    // Send notifications.
    // TODO: implement after registering listeners is implemented.
  }

  return $valid['token'];
}

// Notify others about logging out.
function logout($sid) {
  // Send notifications.
  // TODO: implement after registering listeners is implemented.
}

// Special functions are available only to `$whitelist`-ed IPs.
if (in_array(userip(), $whitelist)) {
  // Check if user with given token is valid (is logged into Ubersmith),
  // and get some basic information about her/him:
  // 1. username
  // 2. e-mail
  // 3. unique ID that also specifies type of user account:
   //   [char TYPE][numeric ID for that type]
  //    `a` = administrator account
  //    `b` = client account
  //    `c` = contact account
  //    Example: `a23` means administrator with account ID = 23.
  // Output is converted to JSON and base64-encoded.
  if (isset($_REQUEST['validate'])) {
    $valid = validate($_REQUEST['validate'], (isset($_REQUEST['ip']) ? $_REQUEST['ip'] : NULL));
    if (empty($valid)) {
      exit(-1);
    }

    $result = array(
      'user' => $valid['username'],
      'uuid' => '',
      'email' => '',
      'fullname' => '',
      //'raw' => $valid,
    );

    if (!empty($valid['admin_id'])) {
      $result['user'] = ($valid['admin_login'] ? $valid['admin_login'] : ($valid['username'] ? $valid['username'] : $valid['admin_id']));
      $result['uuid'] = 'a'.$valid['admin_id'];
      $result['email'] = $valid['admin_email'];
      $result['fullname'] = ($valid['admin_realname'] ? $valid['admin_realname'] : '');
    }
    elseif (!empty($valid['contact_id'])) {
      $result['user'] = ($valid['contact_login'] ? $valid['contact_login'] : ($valid['username'] ? $valid['username'] : $valid['contact_id']));
      $result['uuid'] = 'c'.$valid['contact_id'];
      $result['email'] = $valid['contact_email'];
      $result['fullname'] = ($valid['contact_realname'] ? $valid['contact_realname'] : '');
    }
    elseif (!empty($valid['client_id'])) {
      $result['user'] = ($valid['client_login'] ? $valid['client_login'] : ($valid['username'] ? $valid['username'] : $valid['client_id']));
      $result['uuid'] = 'b'.$valid['client_id'];
      $result['email'] = $valid['client_email'];
      $result['fullname'] = ($valid['client_realname'] ? $valid['client_realname'] : '');
    }

    echo base64_encode(json_encode($result));
    exit();
  }
  // Add our magic to `.htaccess` file.
  elseif (isset($_REQUEST['install'])) {
    $htaccess = file_get_contents('.htaccess');
    if (strpos($htaccess, $magic) === FALSE) {
      // First remove any versions we might have installed before.
      $htaccess = preg_replace('/\n?### START: '.preg_quote($HTAID).' ###[\w\W\s\n]*\n### END: '.preg_quote($HTAID).' ###\n*/', '', $htaccess);
      // Now install our current version.
      $htaccess = "\n{$magic}\n\n{$htaccess}";
      if (file_put_contents('.htaccess', $htaccess) === false) {
        header('HTTP/1.1 503 Service Unavailable');
        echo 'Could not write to .htaccess file';
        exit(-1);
      }
    }
    if (!empty($_REQUEST['install'])) echo $_REQUEST['install'];
    else echo 'Trigger functionality installed.';
    exit();
  }
  // Remove our magic from `.htaccess` file.
  elseif (isset($_REQUEST['uninstall'])) {
    $htaccess = file_get_contents('.htaccess');
    $htaccess = preg_replace('/\n?### START: '.preg_quote($HTAID).' ###[\w\W\s\n]*\n### END: '.preg_quote($HTAID).' ###\n*/', '', $htaccess);
    if (file_put_contents('.htaccess', $htaccess) === false) {
      header('HTTP/1.1 503 Service Unavailable');
      echo 'Could not write to .htaccess file';
      exit(-1);
    }
    if (!empty($_REQUEST['uninstall'])) echo $_REQUEST['uninstall'];
    else echo 'Trigger functionality removed. You can remove this script now.';
    exit();
  }
  // Output our magic for `.htaccess` file. This may be useful to show `mod_rewrite` rules to the administrator.
  elseif (isset($_REQUEST['magic4htaccess'])) {
    echo $magic;
    exit();
  }
}

// Check if Ubersmith's session is still there and matches our token.
if (!isset($_COOKIE[$sname])) {
  // No session, logout.
  if (isset($_COOKIE[$cname])) {
    logout($_COOKIE[$cname]);
    setcookie($cname, '', 0, '/', $domain, false, false);
  }
}
elseif (isset($_COOKIE[$cname])) {
  $valid = login($_COOKIE[$sname]);
  if ($_COOKIE[$cname] != $valid) {
    // Different token means that user logged in to different account, so logout previous one.
    logout($_COOKIE[$cname]);
  }
  setcookie($cname, $valid, 0, '/', $domain, false, false);
}
else {
  $valid = login($_COOKIE[$sname]);
  if (!empty($valid)) {
    setcookie($cname, $valid, 0, '/', $domain, false, false);
  }
}

// Handle `logout.php` and `index.php` overrides, so we can redirect somewhere else than to Ubersmith's login or main page.
if (isset($_REQUEST['redirect'])) {
  setcookie('redirect', $_REQUEST['redirect'], 0, '/', $domain, false, false);
  if ($_GET['q'] == 'logout.php') {
    include('logout.php');
  }
  else {
    include('index.php');
  }
  exit();
}
// When `logout.php` or `index.php` was to be redirected somewhere, our `.htaccess` magic redirects here,
// so we can redirect user to the specified URL (otherwise Ubersmith would redirect to its own pages).
elseif (isset($_REQUEST['redirect_now'])) {
  if (isset($_COOKIE['redirect']) && trim($_COOKIE['redirect']) != '') {
    // Try to prevent redirecting to possibly unsafe URLs (javascript, data, etc...)
    if (preg_match('%^https?://[\w\d\./-_\%#?!&;]+%', $_COOKIE['redirect'])) header('Location: '.$_COOKIE['redirect']);
    setcookie('redirect', '', 0, '/', $domain, false, false);
    exit();
  }
  // No redirect specified, so just run `index.php`.
  else {
    include('index.php');
    exit();
  }
}

// Finally, include or print the file that is used by us to hook into Ubersmith's login/logout.
if (file_exists($hookto)) {
  if ($hookrun) {
    if (basename($hookto) != $hookto) chdir(dirname($hookto));
    include($hookto);
  }
  else readfile($hookto);
}
